#include <stdio.h>

#define ISDEAD 1
#define ISLIVE 0

/*
 * constants for the different bit positions. 
 */

#define M0 1
#define M1 2
#define M2 4
#define M3 8
#define M4 16
#define M5 32
#define M6 64
#define M7 128
#define M8 256
#define M9 512
#define M10 1024
#define M11 2048
#define M12 4096
#define M13 8192
#define M14 16384
#define M15 32768
#define M16 65536
#define M17 131072
#define M18 262144
#define M19 524288
#define M20 1048576
#define M21 2097152
#define M22 4194304
#define M23 8388608
#define M24 16777216
#define M25 33554432
#define M26 67108864
#define M27 134217728
#define M28 268435456
#define M29 536870912

int masks[30]={1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048,
	       4096, 8192, 16384, 32768, 65536, 131072,  262144,
	       524288, 1048576, 2097152, 4194304, 8388608, 16777216,
	       33554432, 67108864, 134217728, 268435456, 536870912};


int arctocode[7][7]={{-1,-1,-1,-1,-1,-1,-1},
                     {-1,-1, 0, 2, 3,12,20},
                     {-1, 1,-1, 4, 5,14,22},
                     {-1, 6, 7,-1,10,18,24},
                     {-1, 8, 9,11,-1,16,26},
                     {-1,13,15,19,17,-1,28},
                     {-1,21,23,25,27,29,-1}};

int movetoarci[30]={1,2,1,1,2,2,3,3,4,4,3,4,
                    1,5,2,5,4,5,3,5,1,6,2,6,3,6,4,6,5,6};
int movetoarcj[30]={2,1,3,4,3,4,1,2,1,2,4,3,
                    5,1,5,2,5,4,5,3,6,1,6,2,6,3,6,4,6,5};

/*
 * The four triangles which end the triangle avoidance game on four nodes.
 */

#define NHLINES 12
#define NPOS 4096

#define TRI1 (M0+M1+M5+M9+M3+M8)
#define TRI2 (M0+M1+M4+M7+M2+M6)
#define TRI3 (M2+M6+M3+M8+M10+M11)
#define TRI4 (M4+M7+M10+M11+M5+M9)

/*
 * Store information about the status of each position and the best move
 * from that position.
 * If winfor[pos]==1, then this is a win for the player to move if he follows
 * the strategy in bestmove[pos] If winfor[pos]==2, then it's a loss.
 */

int winfor[NPOS];
int bestmove[NPOS];
int depth[NPOS];

/*
 * Print out a position in the four node game.
 */

int printpos(pos)
     int pos;
{
  printf("\n");
  printf("1");
  if ((pos & M0) == M0)
    printf("-");
  else
    printf(" ");
  if ((pos & M1) == M1)
    printf("-");
  else
    printf(" ");
  printf("2\n");
  if ((pos & M2) == M2)
    printf("|");
  else
    printf(" ");
  if ((pos & M3) == M3)
    printf("\\");
  else
    printf(" ");
  if ((pos & M4) == M4)
    printf("/");
  else
    printf(" ");
  if ((pos & M5) == M5)
    printf("|");
  else
    printf(" ");
  printf("\n");
  if ((pos & M6) == M6)
    printf("|");
  else
    printf(" ");
  if ((pos & M7) == M7)
    printf("/");
  else
    printf(" ");
  if ((pos & M8) == M8)
    printf("\\");
  else
    printf(" ");
  if ((pos & M9) == M9)
    printf("|");
  else
    printf(" ");
  printf("\n");
  printf("3");
  if ((pos & M10) == M10)
    printf("-");
  else
    printf(" ");
  if ((pos & M11) == M11)
    printf("-");
  else
    printf(" ");
  printf("4");
  printf("\n");
  return(0);
}

/*
 * Check a position to see whether it is live or dead.  This routine
 * is specific to triangle avoidance on four nodes, and would have to
 * be rewritten for five or six nodes or the Hamiltonian Circuit avoidance
 * problem.
 */

int isdead(pos)
     int pos;
{
  if ((pos & TRI1) == TRI1) 
    {
      return(ISDEAD);
    };
  if ((pos & TRI2) == TRI2) 
    {
      return(ISDEAD);
    };
  if ((pos & TRI3) == TRI3) 
    {
      return(ISDEAD);
    };
  if ((pos & TRI4) == TRI4) 
    {
      return(ISDEAD);
    };

  return(ISLIVE);
}

/*
 * This routine should be workable for all of problems.
 */
int checkpos(pos)
     int pos;
{
  int i;
  int newpos;
  int bestlosemove;
  int bestwinmove;
  int bestlosedepth;
  int bestwindepth;
  int nowin;

  nowin=1;

  bestwindepth=100;
  bestlosedepth=0;

  for (i=0; i<=NHLINES-1; i++)
    {
      if ((pos & masks[i]) == 0)
	{
	  newpos=pos | masks[i];

	  if (winfor[newpos] == 0)
	    {
	      checkpos(newpos);
	    }

	  if (winfor[newpos] == 2)
	    {
	      /*
	       * This is a winning move. Check to see if it is quickest.
	       */

	      if (depth[newpos] <= bestwindepth)
		{
		  nowin=0;
		  bestwinmove=i;
		  bestwindepth=1+depth[newpos];
		};

	    }
	  else
	    {
	      /*
	       * This is a losing move.
	       */
	      if (depth[newpos] >= bestlosedepth)
		{
		  bestlosemove=i;
		  bestlosedepth=1+depth[newpos];
		};
	    };
	};
    }

  if (nowin == 1)
    {
      /*
       * We have a losing move.
       */
      bestmove[pos]=bestlosemove;
      winfor[pos]=2;
      depth[pos]=bestlosedepth;
    }
  else
    {
      /*
       * We have a winning move.
       */
      bestmove[pos]=bestwinmove;
      winfor[pos]=1;
      depth[pos]=bestwindepth;
    };


  return(0);
}

int main()
{
  int i;
  int j;
  int move;
  int pos;
  int movebit;
  int first;

  for (i=0; i<=NPOS-1; i++)
    {
      if (isdead(i) == ISDEAD)
	{
	  bestmove[i]=0;
	  winfor[i]=1;
	  depth[i]=0;
	}
      else
	{
	  bestmove[i]=0;
	  winfor[i]=0;
	  depth[i]=0;
	};
    }

  checkpos(0);

  printf("The game is a win for player %d \n",winfor[0]);

 playagain:
  printf("Let's Play!\n");

  printf("Do you want to play first (1) or second (2)?");
  scanf("%d",&first);

  pos=0;
  if (first == 1)
    {
    firstmove:

      printf("Your move:");
      scanf("%d %d",&i,&j);
      
      if ((i < 1) || (i > 4))
	{
	  printf("Invalid Move.  Try Again. \n");
	  goto firstmove;
	};
      
      if ((j < 1) || (j > 4))
	{
	  printf("Invalid Move.  Try Again. \n");
	  goto firstmove;
	};
      
      if (i==j)
	{
	  printf("Invalid Move.  Try Again. \n");
	  goto firstmove;
	};
      
      movebit=masks[arctocode[i][j]];
      if ((movebit & pos)==movebit)
	{
	  printf("Invalid Move!\n");
	  goto firstmove;
	};
      pos=pos | movebit;
      
      printpos(pos);
      
    };

  while (isdead(pos) == ISLIVE)
    {
      move=bestmove[pos];
      printf("My move is %d %d \n",movetoarci[move],movetoarcj[move]);
      movebit=masks[move];
      pos=pos | masks[move];
      printpos(pos);

      if (isdead(pos)==ISDEAD)
	{
	  printf("You Win!\n");
	  goto playagain;
	};

    yourmove:
      printf("Your move:");
      scanf("%d %d",&i,&j);

      if ((i < 1) || (i > 4))
	{
	  printf("Invalid Move.  Try Again. \n");
	  goto yourmove;
	};

      if ((j < 1) || (j > 4))
	{
	  printf("Invalid Move.  Try Again. \n");
	  goto yourmove;
	};
      
      if (i==j)
	{
	  printf("Invalid Move.  Try Again. \n");
	  goto firstmove;
	};

      movebit=masks[arctocode[i][j]];
      if ((movebit & pos)==movebit)
	{
	  printf("Invalid Move!\n");
	  goto yourmove;
	};
      pos=pos | movebit;
      printpos(pos);
    };

  printf("You Lose!\n");
  goto playagain;
}



